// Heatbugs application. Copyright (C) 1996-1999 Santa Fe Institute.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

// The Heatbug Model swarm encapsulates all the objects used in the
// simulated heatbug world itself (but not the user interface objects)

#import "Citizen.h"
#import <space.h> // Grid2d
#import <objectbase/Swarm.h>
#import "WrappedListGrid2d.h"
#import <simtools.h>

id observerSwarm;
extern int epochTime;

@interface ModelSwarm: Swarm
{
  int pplPerCell, numWorkplaces, playerID;			
  int visionRadius;
  int randomSeed;
  int numFeatures, numTraits;
  unsigned long int nofprofiles; 
  int worldXSize, worldYSize;
  int workXSize, workYSize;

  double randomMoveProbability;

  BOOL randomizeCitizenUpdateOrder;
  id modelActions;				  // scheduling data structures
  id modelSchedule;

  id <List> citizenList;				  // list of all the citizens
  id <List> holdingList;
  WrappedListGrid2d *homeGrid;				  // objects representing
  WrappedListGrid2d *workplace;
  id <List> workplaceList;
  id workerNamer;
  id GammaDist1;

  int numberOfInteractionsThisStep, numberOfChangesThisStep;

  FILE *outputFile;
  char outputFileName[50];
  long int ** tabulator;

  BOOL oneAtATime;

  int  sameCounter;
  double  totalSame;
 double  totalSameSquared;

  int  identCounter;
  double  totalIdent;
 double  totalIdentSquared;

  int  acqCounter;    
  double  totalAcquaintance;
double  totalAcquaintanceSquared;


}
+ setObserver: observer;
- getCitizenList;				  // access methods into the

- (id <Grid2d>)getPositionGrid;			  // model swarm. These methods

// 						  // be observed.
- (BOOL) toggleOneAtATime;
- (BOOL)toggleRandomizedOrder;                    // method to toggle the
                                                 // randomization feature
- createCitizenX: (int) x Y: (int) y;

+ createBegin: aZone;				  // extra methods you
- createEnd;					  // provide for Swarms
-initializeRNGs;
-(double) getBetaVariateAlpha1: (double) alpha1 Alpha2: (double) alpha2;

- buildObjects;
- buildActions;
- chooseSubset;

- activateIn: swarmContext;

-(void) resetCounters;

-(void) incrNumberOfInteractionsThisStep;

-(void) incrNumberOfChangesThisStep;
     
-(unsigned int) getNumberOfInteractionsThisStep;

-(unsigned int) getNumberOfChangesThisStep; 


-(void) incrFeature: (int) f Value: (int) amount;
-(void) incrFeature: (int) f Add: (int) amount1 Subtract: (int) amount2;
-(double) getAverageFeature: (int) f;
-(double) getVarianceFeature: (int) f;
-(double) getEntropyFeature: (int) f;
-(double) getMultidimensionalEntropy;
- registerHarmony: (double) same;

- registerIdentical:(double) ident;

- registerAcquaintance: (double) acq;

- (double) getHarmonyAvg;
- (double) getHarmonyVariance;
- (double) getIdenticalAvg;
- (double) getIdenticalVariance;
- (double) getAcquaintanceAvg;
- (double) getAcquaintanceVariance;

-(void) stepReport;

-(void) writeReportDataHeader;
-(void) writeReportData;

@end
