// 
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.


#import <random.h>
#import <space.h>
#import <simtools.h>
#import <misc.h>
#import <time.h>
#import <math.h>
#import "ModelSwarm.h"
#import "MyArguments.h"
#import "ObserverSwarm.h"
#import "Culture.h"
#import "CountHeads.h"
//#include <dmalloc.h>

int epochTime;

static id
makeProbe (id obj, const char *ivarName)
{
  id probe = [VarProbe createBegin: [obj getZone]];
  
  [probe setProbedClass: [obj getClass]];
  [probe setProbedVariable: ivarName];
  return [probe createEnd];
}

static double
getDouble (id obj, const char *ivarName)
{
  id probe = makeProbe (obj, ivarName);
  double ret = [probe probeAsDouble: obj];
  [probe drop];
  return ret;
} 

static int
getInt (id obj, const char *ivarName)
{
  id probe = makeProbe (obj, ivarName);
  int ret = [probe probeAsInt: obj];
  [probe drop];
  return ret;
}


@implementation ModelSwarm

+ setObserver: observer
{
  observerSwarm=observer;
  return self;
}


- getCitizenList
{
  return citizenList;
}


- (WrappedListGrid2d*)getPositionGrid
{
  return homeGrid;
}


- createCitizenX:(int) x Y:(int)  y
{
      Citizen * aCitizen;
      double  personalCondition, threshold;
      
      int offset;
      WrappedListGrid2d * workGrid; 
  
      offset = [uniformIntRand getIntegerWithMin: 0L
                           withMax: (numWorkplaces - 1)];
     
      workGrid= [workplaceList atOffset: offset];
      //printf("workgrid is %d , %d \n", [workGrid getSizeX],[workGrid getSizeY]);

      aCitizen = [Citizen createBegin: self];
      [aCitizen  setWorkGrid: workplace HomeGrid: homeGrid];
      [aCitizen setName: [workerNamer getNewName]];
      aCitizen = [aCitizen createEnd];
      
      // Add the bug to the end of the list.
      // Now initialize the rest of the heatbug's state.
      
 
      [aCitizen setHomeX:  x  Y: y ];


      x=[uniformIntRand
                    getIntegerWithMin: 0L
                    withMax: ([workGrid getSizeX]-1)];
      y=[uniformIntRand getIntegerWithMin: 0L withMax: ([workGrid getSizeY]-1)];
 
      [aCitizen setWorkX:  x  Y: y ];

      [aCitizen init: playerID];
      playerID++;
      personalCondition = [self getBetaVariateAlpha1: .8+(0.015*x)  Alpha2: 2 -(0.015*y) ];
      threshold = [uniformDblRand
                     getDoubleWithMin: .5 withMax: 1];
   

      [aCitizen setPersonalCondition: personalCondition];
      [aCitizen setVision: visionRadius];
      
      [aCitizen setRandomMoveProbability: randomMoveProbability]; 
   
      [citizenList addLast: aCitizen];
    
  return  aCitizen;
}

- (BOOL)toggleRandomizedOrder
{
  randomizeCitizenUpdateOrder = 
    ( (randomizeCitizenUpdateOrder == NO) ? YES : NO );
  return (randomizeCitizenUpdateOrder);
}

- (BOOL) toggleOneAtATime
{
  oneAtATime = 
    ( (oneAtATime == NO) ? YES : NO );
  return (oneAtATime);
}

+ createBegin: aZone
{
  ModelSwarm *obj;
  id <ProbeMap> probeMap;
  int num;
  // First, call our superclass createBegin - the return value is the
  // allocated HeatbugModelSwarm object.

  obj = [super createBegin: aZone];
  
  obj->numWorkplaces = getInt(arguments, "numWorkplaces") ;
  obj->worldXSize = getInt(arguments, "worldXSize");
  obj->worldYSize = getInt(arguments, "worldYSize");

  if( (num= getInt(arguments, "pplPerCell"))!= -1)
  obj->pplPerCell  = num ; 
  else obj->pplPerCell=100;
  obj->workXSize=getInt(arguments, "workXSize");
  obj->workYSize=getInt(arguments, "workYSize");
  obj->numFeatures=getInt(arguments, "numFeatures");
  obj->numTraits = getInt(arguments, "numTraits");
  obj->randomizeCitizenUpdateOrder=YES;
  obj->oneAtATime=NO;


  if( [arguments getVisionRadiusArg]!= -1)
    obj->visionRadius=[arguments getVisionRadiusArg]; 
  else obj->visionRadius = 2;

  obj->randomizeCitizenUpdateOrder = [arguments getToggleRandomOrder];

  obj->oneAtATime = [arguments getToggleOneAtATime]; // 

  obj->randomMoveProbability = getDouble(arguments,"randomMoveProbability");
  
  obj->randomSeed=[arguments getSeedArg]; 

  probeMap = [EmptyProbeMap createBegin: aZone];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

  // Add in a bunch of variables, one per simulation parameter

  [probeMap addProbe: [probeLibrary getProbeForVariable: "pplPerCell"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "visionRadius"
 				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "worldXSize"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "worldYSize"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "numFeatures"
  			     inClass: [self class]  ]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "numTraits"
   				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForMessage: 
                                      "toggleRandomizedOrder" 
                                    inClass: [self class]]];
 [probeMap addProbe: [probeLibrary getProbeForMessage: 
                                      "toggleOneAtATime" 
                                    inClass: [self class]]];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "randomMoveProbability"
				    inClass: [self class]]];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "randomSeed"
				    inClass: [self class]]];
  [probeLibrary setProbeMap: probeMap For: [self class]];
  
  return obj;
}


- createEnd
{
 
  return [super createEnd];
}


- initializeRNGs
{
  if(randomSeed != -1)
  [randomGenerator setStateFromSeed: randomSeed];

  GammaDist1= [GammaDist  create: self  setGenerator: randomGenerator setAlpha: 2 setBeta: 1]; 
  return self;
}


-(double) getBetaVariateAlpha1: (double) alpha1 Alpha2: (double) alpha2
{
   double y1, y2;
   y1=[GammaDist1 getSampleWithAlpha: alpha1  withBeta: 1];
   y2=[GammaDist1 getSampleWithAlpha: alpha2 withBeta: 1];
   return y1/(y1+y2);    
}



- buildObjects
{
  int i, j, k;

  [super buildObjects];
  [self initializeRNGs];
  
  tabulator = [[self getZone] allocBlock: numFeatures * sizeof(long int *)];
  for (i = 0; i < numFeatures; i ++)
    tabulator[i] = [[self getZone] allocBlock: numTraits * sizeof(long int)];

  for(i=0; i< numFeatures; i++) 
    {
      for(j=0; j< numTraits ; j++) 
	{
	  tabulator[i][j]=0;
	}
    }

  sameCounter=0;
  totalSame = 0;

  identCounter=0;
  totalIdent = 0;

  acqCounter=0;    
  totalAcquaintance = 0;


  workerNamer= [UName create: self setBaseName:"agent"];

  homeGrid = [WrappedListGrid2d createNew: self setSizeX: worldXSize Y: worldYSize];
 
  workplaceList = [List create: self];

  for(i=0; i< numWorkplaces; i++)
    {
      workplace=[WrappedListGrid2d createNew: self setSizeX: workXSize Y: workYSize]; 
      [workplaceList addLast: workplace];       
    }

  holdingList = [List create: self];

  citizenList = [List create: self];

  [Citizen setNumFeatures: numFeatures];
  [Citizen setNumTraits: numTraits];
  [Citizen setObserverSwarm: observerSwarm];
  [Citizen setModelSwarm: self];
  [Citizen setPplPerCell: pplPerCell];
  //    x=[uniformIntRand
  //                  getIntegerWithMin: 0L
  //                  withMax: (worldXSize-1)];
  //    y=[uniformIntRand getIntegerWithMin: 0L withMax: (worldYSize-1)];
  
  //this forces a citizen at each location. 
  playerID=0;

 
      for (i = 0; i < worldXSize; i++) 
	{
	  for (j = 0; j < worldYSize; j++)
	    {
	      for (k = 0; k< pplPerCell; k++)
	      [self createCitizenX: i Y: j];
	    }
	}
 
  //create output file
  epochTime= time(0);    
  nofprofiles= pow (numTraits,numFeatures);

  printf("nofprofiles is %lu \n", nofprofiles);  

  if([arguments getRunArg]!= -1)
  sprintf(outputFileName, "DataCulture%d", [arguments getRunArg]);
  else
  sprintf(outputFileName, "DataCulture%d", epochTime);    

  outputFile = fopen (outputFileName,"w+");
  [self writeReportDataHeader];

  [ObjectSaver save: self toFileNamed: "savedobjectmodelswarm"];
 
  printf("Run argument is %d",[arguments getRunArg]);

  return self;
}

// Here is where the model schedule is built, the data structures
// that define the simulation of time in the model. The core is an
// actionGroup that has a list of actions. That's then put in a Schedule.

- buildActions
{
  [super buildActions];
  

  ///////here is the axelrod way:
  modelActions = [ActionGroup create: self]; 

  [modelActions createActionTo: self message: M(resetCounters)];     

if (oneAtATime == YES)
  {
  [modelActions createActionTo: self message: M(chooseSubset)];
  [modelActions createActionForEach: holdingList  message: M(step)];     
  }
else
  {
  //Option 2:
  if (randomizeCitizenUpdateOrder == YES)
    [[modelActions createActionForEach: citizenList message: M(step)] setDefaultOrder: Randomized]; 	 
  else
    {
      [modelActions createActionForEach: citizenList message: M(step)] ;
      [modelActions createActionForEach: citizenList message: M(sayHi)];
    }
  }


  // [modelActions createActionTo: homeGrid message: M(writeValuesToFile)];
  [modelActions createActionTo: self message: M(stepReport)];
 
 
  modelSchedule = [Schedule createBegin: self];
  [modelSchedule setRepeatInterval: 1];
  modelSchedule = [modelSchedule createEnd];
  [modelSchedule at: 0 createAction: modelActions];

  return self;
}

// Now set up the model's activation. swarmContext indicates where
// we're being started in - typically, this model is run as a subswarm
// of an observer swarm.

- chooseSubset
{
  [holdingList removeAll];
  [NSelect select: 1 from: citizenList into: holdingList];
 
  return holdingList;
}

- activateIn: swarmContext
{
  // First, activate ourselves via the superclass activateIn: method.
  // Just pass along the context: the activity library does the right thing.

  [super activateIn: swarmContext];

  // Now activate our own schedule.

  [modelSchedule activateIn: self];

  // Finally, return our activity.

  return [self getActivity];
}


///////////Rick's stuff:


//////////////////////////////////////////////////////////////////
// stepReport
// Ask averager to update itself, and print its results to
// the reportFILE.
-(void) stepReport {
 

         if ( getCurrentTime() % 1 == 0 ) {
                fprintf( outputFile,
                          " %5lu", getCurrentTime() );
                 [self writeReportData];
         }

	 //printf("Variances %f %f %f",[self getHarmonyVariance],[self getIdenticalVariance], [self getAcquaintanceVariance]);


//    for(i=0;i<numFeatures; i++)
//      {
//    printf("Variance feature %d %f \n",i,[self getVarianceFeature: i]);
//    printf("Entropy feature %d %f \n",i,[self getEntropyFeature: i]);
//    [self getMultidimensionalEntropy];
//      }
 }

-(void) writeReportDataHeader 
{
   int i,j;
   char basename[40]; 
   char featurename[20];
   fprintf( outputFile,
  "T Run Seed  IntAct  Change " );
  for(i=0; i<numFeatures; i++)
    {
      sprintf(featurename,"f%d",i);
      fprintf( outputFile, "%s ",featurename);
      for(j=0; j< numTraits; j++)
       {
	 
       sprintf(basename,"f%dT%d",i,j);
       fprintf( outputFile, "%s ", basename);
	}
      fprintf( outputFile, "average%d variance%d relentropy%d ",i,i,i );
    }
  fprintf( outputFile, "totalEntropy acquaint harmony identical \n");
}

-(void) writeReportData 
{
  int i,j;
  fprintf( outputFile,
	   " %d %d  %8d  %8d ",[arguments getRunArg],[arguments getSeedArg],
	    numberOfInteractionsThisStep, 
	   numberOfChangesThisStep 
	   );
  for(i=0; i< numFeatures; i++)
    {
      fprintf( outputFile, "f%d", i);
      for(j=0; j< numTraits; j++)
	fprintf( outputFile, " %ld ", tabulator[i][j]);
        fprintf( outputFile, " %f %f %f ", [self getAverageFeature: i],
	       [self getVarianceFeature:i],[self getEntropyFeature: i]);
    }

  fprintf(outputFile, "%f %f %f %f \n", [self getMultidimensionalEntropy],[self getAcquaintanceAvg], [self getHarmonyAvg],[self getIdenticalAvg] );

  fflush( outputFile );
}

-(void) resetCounters
{ 
   numberOfInteractionsThisStep = 0;
   numberOfChangesThisStep = 0;

#ifndef ONEATATIME   
   sameCounter=0;
   totalSame = 0; totalSameSquared = 0;



   identCounter=0;
   totalIdent = 0;   totalIdentSquared = 0;

   acqCounter=0;    
   totalAcquaintance = 0;totalAcquaintanceSquared = 0;
#endif
} 



-(void) incrNumberOfInteractionsThisStep {
         ++numberOfInteractionsThisStep;
 }

-(void) incrNumberOfChangesThisStep {
          
         ++numberOfChangesThisStep;
 }

-(unsigned int) getNumberOfInteractionsThisStep {
        return numberOfInteractionsThisStep;
 }

-(unsigned int) getNumberOfChangesThisStep {
        return numberOfChangesThisStep;
 }

-(void) incrFeature: (int) f Value: (int) amount {

        ++tabulator[f][amount];
}

-(void) incrFeature: (int) f Add: (int) amount1 Subtract: (int) amount2 {
        --tabulator[f][amount2];
	++tabulator[f][amount1];
	++numberOfChangesThisStep;
	  }

- registerHarmony: (double) same
{
  ++sameCounter;
  totalSame += same;
  totalSameSquared = totalSameSquared + same*same;
  return self;
}

- registerIdentical:(double) ident
{
  ++identCounter;
  totalIdent += ident;
  totalIdentSquared = totalIdentSquared + ident*ident;
   return self;
}

- registerAcquaintance: (double) acq
{
  ++acqCounter;
  totalAcquaintance = totalAcquaintance+ acq;
  totalAcquaintanceSquared = totalAcquaintanceSquared + acq*acq;
  return self;
}

- (double) getHarmonyAvg
{
  return (double) totalSame/(double) sameCounter;
}

- (double) getIdenticalAvg
{
  return (double) totalIdent/ (double) identCounter;
}

- (double) getAcquaintanceAvg
{
   return (double) totalAcquaintance/(double) acqCounter;
}

-(double) getHarmonyVariance
{
  return  ((double) totalSameSquared/(double) sameCounter)  -  (pow( (double) totalSame/(double) sameCounter, 2));
}

-(double) getIdenticalVariance
{
  return  ((double) totalIdentSquared/(double) identCounter)  -  (pow( (double) totalIdent/(double) identCounter, 2));
}

-(double) getAcquaintanceVariance
{
  return  ((double) totalAcquaintanceSquared/(double) acqCounter)  -  (pow( (double) totalAcquaintance/(double) acqCounter, 2));
}

-(double) getAverageFeature: (int) f
{
  int i, weightedSum =0;

  for(i=0; i< numTraits; i++)
    {
      weightedSum += i*tabulator[f][i];
      //printf("i=%d Tabulatorfi  %d\n", i, tabulator[f][i]);
    }
  return (double) weightedSum/(double) [citizenList getCount];
}

-(double) getVarianceFeature: (int) f
{
  int i, weightedSum= 0;
  double average=  [self getAverageFeature: f];
  for(i=0; i< numTraits; i++)
    weightedSum += pow(((double)i - average),2)* tabulator[f][i];
  return (double) weightedSum/(double) [citizenList getCount];
}

-(double) getEntropyFeature: (int) f
{
  int i = 0;
  double entropy = 0; 
  int num=0;

  for(i=0; i < numTraits; i++)
    {
      double p;
      if((num=tabulator[f][i]) !=0)
	{
	  p= (double)num/(double)[citizenList getCount];
	  entropy +=  p*log(p);
	}
    }
  return entropy / log(1.0/(double)numTraits); 
}

-(double) getMultidimensionalEntropy
{
  id tasteMap;
  unsigned long int i , j ;
  unsigned long int k;
  double entropy =0;
  unsigned long int weights[numFeatures];
  id aCitizen;
  id tasteIndex;
  id aScale;
 
  tasteMap=[Map createBegin: self];
  [tasteMap setCompareFunction: compareIntegers];
  tasteMap = [tasteMap createEnd];
   
  for(k=0;k< numFeatures;k++)
    {
      weights[k]= pow(numTraits,numFeatures-1-k); 
    }

  for(j=0; j< (pplPerCell*worldXSize*worldYSize);j++) 
    {
      int indx=0;
      aCitizen=[citizenList atOffset: j];
      for(i=0; i< numFeatures; i++)
	{
	  indx += weights[i] * [aCitizen getFeature: i]; 
	}

      if((aScale= [tasteMap at: (id) indx]) == nil)
	{
	[tasteMap at: (id) indx insert: [[CountHeads createBegin: self] createEnd]];
	//printf("got a nul scale value \n");
	//printf("so I created slot %lu and the value is now %d \n",indx, [[tasteMap at: (id)indx] getCountValue]);
	}
      else
	{
	[aScale addOne];
	//printf("I found a scale value already existed \n");
	//printf("index %lu scaleCount %d \n",indx, [aScale getCountValue]);
	}
    }

  entropy=0;
  tasteIndex=[tasteMap begin: self];  
  while( (aScale=[tasteIndex next])!=nil)
    {
      double p=0;
      p= (double)[aScale getCountValue]/(double)[citizenList getCount] ;
      entropy +=  p*log(p);
    }

  [tasteIndex drop];  
  [tasteMap deleteAll];
  [tasteMap drop];
  return entropy / log(1.0/(double)nofprofiles);      

}

@end
