// Swarm library. Copyright (C) 1996 Santa Fe Institute.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.
#define __USE_FIXED_PROTOTYPES__  // for gcc headers

// with Pietro Terna's changements, se //pt
// with Paul Johnson's changes, se //pj
//#include <dmalloc.h>

#import <misc.h>
#import <string.h>

#import <collections.h>
//pj for CREATEPROBEDISPLAY
#import <simtoolsgui.h>
#import <random.h>   //allows selectRandomMemberX:Y: function

//pt #import "ListShuffler.h"
//pt with 1.3 use the following (ListShuffler is included in simtools)
#import <simtools.h>

#import "WrappedListGrid2d.h"		// Multiple Occupancy Discrete2d grid

// When an object is added to an empty cell on the grid, a List is created
// at that cell to hold the object. Further additions of objects to the
// cell are just added to the list.

//pj: now we are creating a Cell at that spot, and the list is inside the cell.
//pj: The cell is called ListWrapper  for lack of a better name!

#import "ListWrapper.h"

// Whenever objects are removed from the list, if the list becomes empty
// it is dropped (removed).

//pj: I've implemented recycling here.

// The code for Lists allow for adding arbitrary data values rather than
// object id's, and the code for this grid allows for that.

@implementation WrappedListGrid2d

// Private methods:

-makeOffsets {
  int i;

  // precalculate offsets based on the y coordinate. This lets
  // us avoid arbitrary multiplication in array lookup.
  offsets = [[self getZone] alloc: ysize * sizeof(*offsets)];

  for (i = 0; i < ysize; i++)
    offsets[i] = xsize * i;		// cache this multiplication
  return self;
}

-(id *) allocLattice {
  void * p;
  p = [[self getZone] alloc: xsize * ysize * sizeof(id)];
  memset(p, 0, xsize * ysize * sizeof (id));
  return p;
}

-setSizeX: (int) x Y: (int) y {
  if (lattice)
    [InvalidArgument raiseEvent: 
    "WrappedListGrid2d: You cannot reset the grid size after creation.\n"];
  xsize = x;
  ysize = y;
  return self;
}

-createEnd {
  if (xsize <= 0 || ysize <= 0)
    [InvalidCombination raiseEvent: 
    "MoGrid2d: Invalid size in creation: %d %d\n", xsize, ysize];
  lattice = [self allocLattice];
  [self makeOffsets];
  shuffler = [ListShuffler create: [self getZone]];

  recyclerList = [List create: [self getZone]];

  return self;
}

// Creation and parameter setting:

+createNew: (id) aZone setSizeX: (int) x Y: (int) y {
   WrappedListGrid2d * grid;

   grid = [WrappedListGrid2d createBegin: aZone];
   [grid setSizeX: (int) x Y: (int) y];
   grid = [grid createEnd];

   return grid;
}

// Accessing data values:

-(unsigned) getSizeX {
  return xsize;
}

-(unsigned) getSizeY {
  return ysize;
}

-(id *) getLattice {
  return lattice;
}

//pt -(int *) getOffsets {
-(long *) getOffsets {  //pt long with Grid2d inheritance
  return offsets;
 }

//pj: Add objects to lists within holding cell objects:

-addFirst: (id) anObject atX: (int) x Y: (int) y 
{
  id cell;
//  id <List> list;

   cell = *mo2dSiteAt(lattice, offsets, x, y);

   if (!cell) 
    {						// site is empty now
     cell= [self createCellX: x Y: y];
    }
 
    [[cell getList] addFirst: anObject];		// add object to list
   
    return self;
    }

-pushObject: (id) anObject atX: (int) x Y: (int) y {
  id cell;
 // id <List> list;

  cell = *mo2dSiteAt(lattice, offsets, x, y);
  if (!cell) 	 		                        // site is empty now
    {
   cell= [self createCellX: x Y: y];
    }
   [[cell getList] addFirst: anObject];			// add object to list
     
  return self;
}

-addLast: (id) anObject atX: (int) x Y: (int) y {
  id cell;
 // id <List> list;

  cell = *mo2dSiteAt(lattice, offsets, x, y);

  if (!cell) 
    {						// site is empty now
     cell= [self createCellX: x Y: y];  
     }

  [[cell getList] addLast: anObject];			// add object to list
 
   return self;
}

//pj: eliminated method enQueObject  due to redundancy with previous method
// -enQueObject: (id) anObject atX: (int) x Y: (int) y {
//   id list;

// Remove objects from lists at cells:

-(id) getFirstAtX: (int) x Y: (int) y {
   id cell,item;
   id <List> list;
   int num;

   cell = *mo2dSiteAt(lattice, offsets, x, y);

   if (!cell) return cell;	// i.e. nil; cell has no list
 
   list=[cell getList];
   num = [ list getCount];
   
   if (num > 1)
     return [list removeFirst];

   if (num == 1) {
     item = [list removeFirst];
     // Now list should be empty; remove list
     [recyclerList addLast: cell];
     *mo2dSiteAt(lattice, offsets, x, y) = nil;	       // nullify the Grid value
     return item;  
   }

   // If we're here, num = 0 which should not happen!

   [recyclerList addLast: cell];
      *mo2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
     printf("WrappedListGrid2d: deleteFirst: found empty list!\n");
     return nil;

}

-(id) popObjectAtX: (int) x Y: (int) y {
   id cell,item;
   id <List> list;
   int num;

   cell = *mo2dSiteAt(lattice, offsets, x, y);
   
   if (!cell) return nil;	// i.e. nil; cell has no list
 
   list=[cell getList];
   num = [list getCount];
   if (num > 1)
     return [list removeFirst];
   if (num == 1) {
     item = [list removeFirst];
     [recyclerList addLast: cell];
      *mo2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
     return item;  
   }
// If we're here, num = 0 which should not happen!
      [recyclerList addLast: cell];
      *mo2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
     printf("WrappedListGrid2d: deleteFirst: found empty list!\n");
     return nil;
}


//pj: redundant with previous method, so I eliminated this one:
// -(id) deQueObjectAtX: (int) x Y: (int) y {


-(id) getLastAtX: (int) x Y: (int) y {
   id cell,item;
   id <List> list;
   int num;

   cell = *mo2dSiteAt(lattice, offsets, x, y);

   if (!cell) return nil;	// i.e. nil; cell has no list
 
   list=[cell getList];
   num = [list getCount];
   
   if (num > 1)
     return [list removeLast];

   if (num == 1) {
     item = [list removeLast];
      [recyclerList addLast: cell];
      *mo2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
     return item;  
   }

   // If we're here, num = 0 which should not happen!
     [recyclerList addLast: cell];
     *mo2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
     printf("MoGrid2d: deleteFirst: found empty list!\n");
     return nil;

}

//Note: this does not delete or remove the randomly chosen member.
//it simply gives back a pointer to it.

-(id) selectRandomMemberX: (int) x Y: (int) y
{
	id cell,item = nil;
	id <List> list;
   int num;

   cell = *mo2dSiteAt(lattice, offsets, x, y);
   list=[cell getList];
   num = [list getCount];
   if (num ==0) return nil;
   if (num ==1) return [list atOffset:0];
   if (num > 1)
     {
     item= [list atOffset: [uniformUnsRand getUnsignedWithMin: 0 withMax: num-1]];
     }
   return item;  
}

-(id) deleteObject: (id) anObject atX: (int) x Y: (int) y {
   id <List> list;
   id cell;
   id item;

   cell = *mo2dSiteAt(lattice, offsets, x, y);
   if (!cell) return nil;	// no list there, so return nil

   list=[cell getList];
   if(!list)    [InvalidArgument  raiseEvent: "You got a bogus spot in positionGrid \n"];

   item = [list remove: anObject]; 
   if ([list getCount]==0) 
     {
       [recyclerList addLast: cell];
       *mo2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
     }
   return item;
}


-(void) dropListAtX: (int) x Y: (int) y {
   id cell;
   
   cell = *mo2dSiteAt(lattice, offsets, x, y);
   if (cell) {
     [recyclerList addLast: cell];
    
     *mo2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
   }
   // no return
}

-(void) dropAllAtX: (int) x Y: (int) y {
   id cell;
   id <List> list;

   cell = *mo2dSiteAt(lattice, offsets, x, y);
  
   if (cell) {
     list=[cell getList];
     [list forEach: M(drop)];
     [recyclerList addLast: cell]; 
     *mo2dSiteAt(lattice, offsets, x, y) = nil;	// zero the cell value
   }
   // no return
}

// Retrieve object data from cells:

-(int) numberOfObjectsAtX: (int) x Y: (int) y {
   id <List> list;
   id cell;
  
   cell = *mo2dSiteAt(lattice, offsets, x, y);  
   list=[cell getList];
   return [list getCount]; 
}


-getListAtX: (int) x Y: (int) y 
{
  id cell;
  cell=*mo2dSiteAt(lattice, offsets, x, y);
  return [cell getList];
}



-getCopyAtX: (int) x Y: (int) y 
{
   id <List> list;
   id <List>  newlist;
   id cell;
   int i, num;

   cell = *mo2dSiteAt(lattice, offsets, x, y);
   if (!cell) return cell;

   list=[cell getList];
   num = [list getCount];
   if (num == 0) return nil;

   // copy the list:
   newlist = [List create: [self getZone]];
   for (i=0; i<num; i++)
     [newlist addLast: [list atOffset: i]];

   return newlist;
}


// Retrieve randomized lists from cells:

-(id) getShuffledListAtX: (int) x Y: (int) y {
   id <List> list;
   id cell;

   cell = *mo2dSiteAt(lattice, offsets, x, y);
   if (!cell) return cell;	// nil, i.e. no list

   list=[cell getList];
   if ([list getCount] == 0)	// Should Not Happen
     return nil;
   else {
     [shuffler shuffleWholeList: list];
     return list;
   }
}


-getShuffledCopyAtX: (int) x Y: (int) y {
   id <List> list, newlist;
   id cell;
   int num,i;


   cell = *mo2dSiteAt(lattice, offsets, x, y);
   if (!cell) return cell;	// nil, i.e. no list
 
   list=[cell getList];
   if ([list getCount] == 0)	// Should Not Happen
     return nil;
   else {
     newlist = [List create: [self getZone]];
     num = [list getCount];
     for (i=0; i<num; i++)
       [newlist addLast: [list atOffset: i]];
     return [shuffler shuffleWholeList: newlist];
   }
}


-(void) shuffleListAtX: (int) x Y: (int) y {
   id cell;

   cell = *mo2dSiteAt(lattice, offsets, x, y);
   if (!cell) return;

   [shuffler shuffleWholeList: [cell getList]];
}

-(void) shuffleAll {
   id cell;
//   id <List> list;
   int ii,jj;

   for  (ii=0; ii<xsize; ii++) {
   for  (jj=0; jj<ysize; jj++) {
      cell = *mo2dSiteAt(lattice, offsets, ii, jj);
      if (cell) [shuffler shuffleWholeList: [cell getList]];
   } }
}

//this deletes the list that used to be in the Wrapper.
//puts new one in there.
//I don't know why this would be used, but Sven had it, so it must be useful!
// It checks if the existing list is empty.

-setNewList: (id) newlist atX: (int) x Y: (int) y {
//   id <List> list;
   id cell;

   cell = *mo2dSiteAt(lattice, offsets, x, y);
   if (!cell) 
    {						// site is empty now
     cell= [self createCellX: x Y: y];  
     }

   if ([[cell getList] getCount]!=0) 
   [InvalidCombination raiseEvent:
   "WrappedListGrid2d: setList: atX: Y:  (x,y) already occupied!\n"];
   [cell setNewList: newlist];
   return self;
}

// pj got tired of updating cell creation code in so many places
// so it is all here in one method

-createCellX: (int) x Y: (int) y
{
  id <List> list;
  id cell;

  if ( ([recyclerList getCount]!=0) ) 
     {
       cell=[recyclerList removeFirst];
       
     }
 else
    {
    cell= [ ApplicationSpecific  createBegin: [self getZone]];
    cell=[cell createEnd];
    list = [List create: [self getZone]];		// create empty cell
    [cell setList: list];              // put list in cell
       
     }
    [cell setX: x Y: y];	
    [cell initializeAllValues];
    *mo2dSiteAt(lattice, offsets, x, y) = cell; 
  return cell;
}


//pj inserted to grab random  item in list at x y
- makeProbeAtX: (unsigned)x Y: (unsigned)y
{
  id obj;
  id cell;
  id <List> list;

  if (x >= 0
      && x < xsize
      && y >= 0
      && y < ysize)
    {
      cell = *mo2dSiteAt(lattice, offsets, x, y);
      list=[cell getList];
      obj = [list  atOffset: [uniformUnsRand getUnsignedWithMin: 0 withMax: [list getCount]-1] ];
      if (obj)
	{
        CREATE_PROBE_DISPLAY (obj);
       	}
      else
        GUI_BEEP ();
    }
    else
      [WarningMessage
        raiseEvent:
          "Object2dDisplay: invalid coordinates to make probe (%d,%d)\n",
        x, y];
  return self;
}

//inserted to make this write itself to a file!

-writeValuesToFile
{
   id gridfiler;
   char filename[40];
  sprintf(filename, "%05ldWrappedGrid.txt", getCurrentTime());
   gridfiler=[Int2dFiler createBegin: [self getZone]];
   [gridfiler setDiscrete2dToFile: self];
   [gridfiler setValueMessage: M(getOutputValue)];
   gridfiler=[gridfiler createEnd];
    
   [gridfiler fileTo: filename];
   return self;
}


@end
